/* ======================================================================== *
				spacer.c - Subroutines for laying out gadgets.
							       By Talin.
				Compiled under Manx 5.0 with large code/data
 * ======================================================================== *

		'spacer.c' is a set of subroutines for laying out rectangular regions
	in a visually attractive way. It has a number of modes and options which
	allow, for example, gadgets to be laid out in a font-sensitive fashion
	and look nice.
		Note that these are fairly low-level routines that perform the basic
	calculations for figuring out gadget spacing. You would probably want a
	couple of higher level layout routines on top of these in a real
	application.

		The first routine, "ArrayTextLength" is a very simple one. You pass
	to it an array of pointers to text strings, as well as a font pointer,
	and it fills in another array with the pixel-widths of all those strings.
	In addition, it also returns the pixel-width of the longest text string.
		This routine is usually used to figure out text lengths for font-
	sensitive gadget layout. It is a very short subroutine, and is only
	placed here for convenience.
		The calling parameters are as follows:

			longest = ArrayTextLength(
				char *textarray[],
				UWORD resultarray[],
				int count,
				struct TextFont *font);

		textarray is the array of pointer to text.
		resultarray is the array of UWORD results.
		count is the length of both arrays.
		font is the font pointer.

		The second routine, "ArrangeSpacing" actually performs two seperate
	functions, depending upon the parameters it is called with. However, before
	I discuss that, I will need to explain the basic layout strategy I am using.

		In all cases, we start with an array of "element widths". This could be,
	for example, the text string lengths as provided by ArrayTextLength, above.
	Now, we don't want each element to be jammed right up next to the other
	one, so we need to put some space in between the elements. This can be done
	several ways:
		1. We can put a fixed amount of space between each element, and then
			make the container area big enough to hold all the elements plus
			the available width.
		2. If dealing with a container of previously fixed size, we can
			put a fixed amount of space between each element, and center all
			the elements within the container.
		3. Instead of #2, we can subtract the sum of all the element widths from
			the available space, and then divide it up between the elements so
			that they are "spread out" across the container. In other words,
			the space between the gadgets "adapts" to the larger container.

		Now, usually you don't just have a naked text string for a button, there
	is usually some kind of decorative border around it. Now, there needs to
	be some space between the edges of the text string and the border, and there
	needs to be some space between the border for one gadget and the next one.
		For purposes of this discussion, we shall call the space between the
	gadgets the "interval size" and the space between the text and the border
	the "border size". So, here is a drawing:

          +-------------------+      +-------------------+      
          |                   |      |                   |
          |      Label<---+-->|<--+->|       Label       |
          |               |   |   |  |                   |
          +---------------+---+   |  +-------------------+      
                          |       +-interval
                          +-border

		Now, the number of options gets even larger. The border can be "adaptable",
	i.e. change based on the container size, or the interval can be adaptable,
	or both.

		In addition, it is to be noted that when you have a lot of buttons that
	are nearly the same size, it looks nicer if they are exactly the same
	size.

		Now let's consider the spacing around the group of gadgets as a whole.
	I have simplified this to four cases:
		1. The outside interval can be zero.
		2. The outside interval can be half the normal interval.
		3. The outside interval can be the same as the normal interval.
		4. The outside interval can be twice the normal interval.

		OK, now let's talk about the subroutine ArrangeSpacing. This can be called
	in two ways: The first way is when you want to know how big of a container
	to make, and the second is when you want to lay out gadgets within the
	container. Which way the routine works is based on whether or not the
	'offsetarray' parameter is NULL or not.
		Note that ArrangeSpacing is ONE-DIMENSIONAL -- if you want to use it for
	an array of elements, rather than just a row or column, you will need to call
	it twice -- one for the X dimension and once for the Y.

		In all cases, the routine takes some flags:

			FIXED_BORDER is a flag which means "do not adapt the border size",
			FIXED_INTERVAL is a flag which means "do not adapt the interval".

		The calling parameters are as follows:

			int ArrangeSpacing(
				int					total_space,
				UWORD				lengtharray[],
				UWORD				offsetarray[],
				int					count,
				int					min_border,
				int					min_interval,
				int					out_type,
				int					samesize,
				int					flags)
		
		total_space is the size of the container, or 0 if doing a size query.
		lengtharray is the length of each element, and be the result array
			from ArrayTextLength. If the 'offsetarray' parameter (below) is
			non-NULL, then the elements in this array will be modified to
			represent the "adapted" sizes of the elements.
		offsetarray can be NULL, which means that you only want to know
			the minimum container size, or it can point to an array of 'count'
			UWORDS, in which case it will be filled in by this routine. Each
			element of the array will indicate how many pixels from the
			container edge to draw the element.
		min_border is the minimum permissable border size. If FIXED_BORDER
			is set, this WILL be the border size.
		min_interval is the minimum permissable interval size. If FIXED_INTEVAL
			is set, this WILL be the interval size.
		out_type represents the code indicating how large the outside interval
			should be in relation to the standard interval. It's value can be:

				NO_OUTSIDE_INTERVAL
				HALF_OUTSIDE_INTERVAL
				FULL_OUTSIDE_INTERVAL
				DOUBLE_OUTSIDE_INTERVAL

		samesize is a BOOLEAN flag which indicates that you want the elements
			to all be forced to the size of the largest one.
		flags is the two flags FIXED_BORDER and FIXED_INTERVAL, as mentioned
			above.

		The value returned is either the size of the container (if offsetarray
	is NULL) or 0.

 * ======================================================================== */

#include <std_headers.h>
#include "spacer.h"

int ArrayTextLength(char **, UWORD *, int, struct TextFont *);
int ArrangeSpacing(int, UWORD *, UWORD *, int, int, int, int, int, int);

#if 0

; register usage
;	d0 - scratch
;	d1 - scratch
;	d2 - number of array elements to do
;	d3 - font pointer
;	d4 = longest

;	a0 - scratch register, and string address
;	a1 - scratch
;	a2 - scratch
;	a3 - text array
;	a4 - pointer to temporary rastport
;	a5 - result array

			xref	_GfxBase

atregs		reg		d2-d4/a2-a5					; 7 registers

_ArrayTextLength:
			movem.l	atregs,-(sp)

			move.l	32(sp),a3					; address of text array
			move.l	36(sp),a5					; address of result array
			move.l	40(sp),d2					; number of elements in array
			move.l	44(sp),d3					; font pointer

			move.l	_GfxBase,a6					; get GfxBase while a4 still good

			sub.l	#rp_sizeof,sp				; alloc space for temp rastport
			move.l	sp,a4
			move.l	sp,a1						; a1 <-- rastport pointer
			Call	InitRastPort				; initialize the rastport

			move.l	a4,a1						; a1 <-- rastport pointer
			move.l	d3,a0						; a0 <-- font pointer
			Call	SetFont						; set rastport's font pointer

			moveq	#0,d4						; longest = 0

10$			subq.w	#1,d2						; subtract 1 from array size count
			bmi.s	90$							; if negative, then quit

			move.l	(a3)+,a0					; a0 <-- address of string #n
			move.l	a0,a2						; a2 <-- start of string #n

			moveq	#-1,d0						; initialize string length d0 to -1
20$			addq.l	#1,d0						; add 1 to length
			tst.b	(a2)+						; if end of string
			bne.s	20$							; then quit, else loop back

			move.l	a4,a1						; a1 <-- temp rastport
			Call	TextLength					; find length of string in pixels

			move.w	d0,(a5)+					; store length in result array

			cmp.w	d0,d4						; if longest > length
			bgt.s	30$							; then skip
			move.w	d0,d4						; else longest = length
30$
			bra.s	10$

90$			moveq	#0,d0
			move.w	??,d0
			movem.l	(sp)+,atregs
			rts


#endif


	/* finds the amount of space needed by each text string, as well as
		the longest text string.
	*/

int ArrayTextLength(
	char				*textarray[],
	UWORD				resultarray[],
	int					count,
	struct TextFont		*font)
{	struct RastPort		temp_rp;
	int					longest=0;

	InitRastPort(&temp_rp);
	SetFont(&temp_rp,font);

	while (count--)
	{	*resultarray = TextLength(&temp_rp,(UBYTE *)(*textarray),strlen(*textarray));
		if (*resultarray > longest)
			longest = *resultarray;
		resultarray++;
		textarray++;
	}
	return longest;
}

int ArrangeSpacing(
	int					total_space,
	UWORD				lengtharray[],
	UWORD				offsetarray[],
	int					count,
	int					min_border,
	int					min_interval,
	int					out_type,
	int					samesize,
	int					flags)
{	WORD				numgaps = count+1,
						numintervals,
						interval;
	WORD				largest_size=0,
						sum_size=0;
	WORD				i;
	int					requiredspace;
	int					divy;
	int					offset = 0;
	int					portions,
						bonus;

	if (count < 1) return 0;

	if (out_type == NO_OUTSIDE_INTERVAL) numgaps-=2;
	else if (out_type == HALF_OUTSIDE_INTERVAL) numgaps--;
	else if (out_type == DOUBLE_OUTSIDE_INTERVAL) numgaps += 2;

		/* calculate the sum size as well as the largest size */

	for (i=0; i<count; i++)
	{	sum_size += lengtharray[i];
		if (lengtharray[i] > largest_size) largest_size = lengtharray[i];
	}

	requiredspace = count * min_border * 2 + numgaps * min_interval;
	
	if (samesize) requiredspace += count * largest_size;
	else requiredspace += sum_size;

	if (offsetarray == NULL) return requiredspace;

	if (flags == 0) numintervals = numgaps + count*2;
	else if (flags == FIXED_INTERVAL) numintervals = count*2;
	else if (flags == FIXED_OUTSIDE_INTERVAL) numintervals = numgaps;

	interval = 0;

	divy = total_space - requiredspace;

	if (out_type == NO_OUTSIDE_INTERVAL)
	{	offset = 0;
		portions = 0;
	}
	else if (out_type == HALF_OUTSIDE_INTERVAL)
	{	offset = min_interval / 2;
		portions = 1;
	}
	else if (out_type == FULL_OUTSIDE_INTERVAL)
	{	offset = min_interval;
		portions = 2;
	}
	else if (out_type == DOUBLE_OUTSIDE_INTERVAL)
	{	offset = min_interval * 2;
		portions = 4;
	}

	if (flags != FIXED_INTERVAL && numintervals > 0)
	{	bonus = divy * portions / numintervals;
		numintervals -= portions;
		offset += bonus / 2;
		divy -= bonus;
	}

	for (i=0; i<count; i++)
	{	offsetarray[i] = offset;

		if (samesize) lengtharray[i] = largest_size;

		if (flags != FIXED_OUTSIDE_INTERVAL && numintervals > 0)
		{	bonus = divy * 2 / numintervals;
			numintervals -= 2;
			divy -= bonus;
			lengtharray[i] += bonus + min_border * 2;
		}

		offset += lengtharray[i] + min_interval;

		if (flags != FIXED_INTERVAL && numintervals > 0)
		{	bonus = divy / numintervals;
			numintervals--;
			offset += bonus;
			divy -= bonus;
		}
	}
	return 0;	/* what to return here... */
}

#if 0

asregs		reg		??

_ArrangeSpacing:

int ArrangeSpacing(
	int					total_space,
	UWORD				lengtharray[],
	UWORD				offsetarray[],
	int					count,
	int					min_border,
	int					min_interval,
	int					out_type,
	int					samesize,
	int					flags)
{	WORD				numgaps = count+1,
						numintervals,
						interval;
	WORD				largest_size=0,
						sum_size=0;
	WORD				i;
	int					requiredspace;
	int					divy;
	int					offset = 0;
	int					portions,
						bonus;

	if (count < 1) return 0;

	if (out_type == NO_OUTSIDE_INTERVAL) numgaps-=2;
	else if (out_type == HALF_OUTSIDE_INTERVAL) numgaps--;
	else if (out_type == DOUBLE_OUTSIDE_INTERVAL) numgaps += 2;

		/* calculate the sum size as well as the largest size */

	for (i=0; i<count; i++)
	{	sum_size += lengtharray[i];
		if (lengtharray[i] > largest_size) largest_size = lengtharray[i];
	}

	requiredspace = count * min_border * 2 + numgaps * min_interval;
	
	if (samesize) requiredspace += count * largest_size;
	else requiredspace += sum_size;

	if (offsetarray == NULL) return requiredspace;

	if (flags == 0) numintervals = numgaps + count*2;
	else if (flags == FIXED_INTERVAL) numintervals = count*2;
	else if (flags == FIXED_OUTSIDE_INTERVAL) numintervals = numgaps;

	interval = 0;

	divy = total_space - requiredspace;

	if (out_type == NO_OUTSIDE_INTERVAL)
	{	offset = 0;
		portions = 0;
	}
	else if (out_type == HALF_OUTSIDE_INTERVAL)
	{	offset = min_interval / 2;
		portions = 1;
	}
	else if (out_type == FULL_OUTSIDE_INTERVAL)
	{	offset = min_interval;
		portions = 2;
	}
	else if (out_type == DOUBLE_OUTSIDE_INTERVAL)
	{	offset = min_interval * 2;
		portions = 4;
	}

	if (flags != FIXED_INTERVAL && numintervals > 0)
	{	bonus = divy * portions / numintervals;
		numintervals -= portions;
		offset += bonus / 2;
		divy -= bonus;
	}

	for (i=0; i<count; i++)
	{	offsetarray[i] = offset;

		if (samesize) lengtharray[i] = largest_size;

		if (flags != FIXED_OUTSIDE_INTERVAL && numintervals > 0)
		{	bonus = divy * 2 / numintervals;
			numintervals -= 2;
			divy -= bonus;
			lengtharray[i] += bonus + min_border * 2;
		}

		offset += lengtharray[i] + min_interval;

		if (flags != FIXED_INTERVAL && numintervals > 0)
		{	bonus = divy / numintervals;
			numintervals--;
			offset += bonus;
			divy -= bonus;
		}
	}
	return 0;	/* what to return here... */
}
#endif
